﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
FrameRateControl v0.3 by MacGruber.
This plugin allows more fine control over VaM's VSync setting and as an alternative you can set a
TargetFrameRate. Additionally you can make the game pause automatically when VaM you are working
in other applications to reduce CPU/GPU load to almost zero.

Version 0.3 2020-03-21
	Adding to MacGruber-Toolkit Package.
	Improved info texts.
	
Version 0.2 2020-01-15
	Fixed issues with VaM 1.18.1.6.
	Fixed issues with not applying settings on load.
	Reduced log spam.
	
Version 0.1 2019-09-22
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;


namespace MacGruber
{	
	public class FrameRateControl : MVRScript
	{
		JSONStorableBool pauseInBackground;
		JSONStorableBool keepSettingsOverSceneChange;
		JSONStorableStringChooser frameRateMode;
		JSONStorableStringChooser vsyncMode;
		
		private readonly List<string> frameRateModes = new List<string>() { "Max", "90", "60", "45", "30", "20" };
		private readonly int[] frameRates = new int[] { -1, 90, 60, 45, 30, 20 };
		private readonly List<string> vsyncModes = new List<string>() { "Off", "On", "Half" };
		private int frameRateModeIdx = 0;
		private int vsyncCount = 1;	
		private bool gameIsVR = false;
		private bool haveInit = false;
		private bool suppressApply = false;
		private bool needApply = true;
		
		public override void Init()
		{
			haveInit = false;
			needApply = true;
			suppressApply = false;
			gameIsVR = XRSettings.isDeviceActive && !string.IsNullOrEmpty(XRSettings.loadedDeviceName);
			if (gameIsVR)
			{
				SetupInfoText(this, 
					"<color=#606060><size=40><b>FrameRateControl v0.3</b></size>\nThis plugin allows more fine control over VaM's VSync settings and framerate. Especially recommended when recording video or you want to reduce power usage or fan noise when working with VaM.</color>\n\n" +
					"<color=#FF0000><b>NOTE:</b> FrameRateControl plugin is disabled in VR.</color>",
					500.0f, true
				);
				return;
			}		
			
			frameRateMode = new JSONStorableStringChooser("TargetFrameRate", frameRateModes, frameRateModes[frameRateModeIdx], "FrameRateTarget");
			frameRateMode.setCallbackFunction += (string modeName) => {
				frameRateModeIdx = frameRateModes.FindIndex((string entry) => { return entry == modeName; });
				if (frameRateModeIdx < 0)
					frameRateModeIdx = 0;
				if (frameRateModeIdx != 0)
				{
					vsyncCount = 0;
					vsyncMode.valNoCallback = vsyncModes[vsyncCount];
				}
				Apply();
			};
			CreateScrollablePopup(frameRateMode, false);
			RegisterStringChooser(frameRateMode);	
			
			vsyncCount = UserPreferences.singleton.desktopVsync ? 1 : 0;
			vsyncMode = new JSONStorableStringChooser("VSync Mode", vsyncModes, vsyncModes[vsyncCount], "VSync Mode");
			vsyncMode.setCallbackFunction += (string modeName) => {
				vsyncCount = vsyncModes.FindIndex((string entry) => { return entry == modeName; });		
				if (vsyncCount < 0)
					vsyncCount = 0;
				if (vsyncCount != 0)
				{
					frameRateModeIdx = 0;
					frameRateMode.valNoCallback = frameRateModes[frameRateModeIdx];
				}
				Apply();
			};		
			
			CreateScrollablePopup(vsyncMode, false);
			RegisterStringChooser(vsyncMode);
			
			pauseInBackground = SetupToggle("Pause VaM when not focused", true, false);
			pauseInBackground.setCallbackFunction += (bool v) => { Apply();	};
			keepSettingsOverSceneChange = SetupToggle("Keep settings over scene change", false, false);
			
			
			SetupInfoText(this, 
				"<color=#606060><size=40><b>FrameRateControl v0.3</b></size>\nThis plugin allows more fine control over VaM's VSync settings and framerate. Especially recommended when recording video or you want to reduce power usage or fan noise when working with VaM.</color>\n\n" +
				"<color=#FF0000><b>WARNING:</b> Physics might get out of sync with time. You might need to tweak your Physics Rate and Physics Update Cap settings until you get satisfiying results.</color>",
				500.0f, true
			);
			
			haveInit = true;			
			StartCoroutine(SyncSettings());
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			suppressApply = true;
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
			suppressApply = false;
			Apply();
		}
		
		private void OnEnable()
		{			
			if (haveInit)
				Apply();
		}
		
		private void Update()
		{
			if (needApply)
				Apply();
		}
			
		private void OnDisable()
		{
			if (gameIsVR)
				return;
			
			if (keepSettingsOverSceneChange.val)
			{
				SuperController.LogMessage("FrameRateControl: Keeping settings over scene change.");
			}
			else
			{		
				Application.runInBackground = true;
				Application.targetFrameRate = -1;
				QualitySettings.vSyncCount = UserPreferences.singleton.desktopVsync ? 1 : 0;			
				SuperController.LogMessage("FrameRateControl: VaM UserPreferences restored");
			}
		}
		
		private void Apply()
		{
			if (gameIsVR || suppressApply)
				return;
			
			Application.runInBackground = !pauseInBackground.val;
			Application.targetFrameRate = frameRates[frameRateModeIdx];
			UserPreferences.singleton.desktopVsync = vsyncCount > 0;
			QualitySettings.vSyncCount = vsyncCount;
				
			string msg = "PauseInBackground " + (pauseInBackground.val ? "On" : "Off");
			if (vsyncCount > 0)
				SuperController.LogMessage("FrameRateControl: VSync " + vsyncModes[vsyncCount] + ", " + msg);
			else
				SuperController.LogMessage("FrameRateControl: Targeted " + frameRateModes[frameRateModeIdx] + " fps, " + msg);
			
			needApply = false;
		}
		
		private IEnumerator SyncSettings()
		{
			while (true)
			{
				// uneven wait time to avoid overlap with other regular tasks
				yield return new WaitForSeconds(1.93f);
				
				// keep VSync setting synced with VaM UserPreferences in case the user changes it there.
				int actualVSyncCount = QualitySettings.vSyncCount;
				if (actualVSyncCount != vsyncCount)
				{
					actualVSyncCount = Mathf.Clamp(actualVSyncCount, 0, 2);
					vsyncMode.val = vsyncModes[actualVSyncCount];
				}
			}
		}
		
		private JSONStorableFloat SetupSliderFloat(string name, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(name, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateSlider(storable, rightSide);
			RegisterFloat(storable);
			return storable;
		}
		
		private JSONStorableBool SetupToggle(string name, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(name, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateToggle(storable, rightSide);
			RegisterBool(storable);
			return storable;
		}
		
		private static JSONStorableString SetupInfoText(MVRScript script, string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = script.CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
	}
}