﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
SkyMagic Loader v0.1 by MacGruber.
Retrieves a cubemap texture from a ReflectionProbe inside your CustomUnityAsset and sets it as source for VaM's Global Sky and Global Illumination.

Version 0.1 2020-04-12
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.Events;
using System.Collections;
using mset;
using System;

namespace MacGruber
{
    public class SkyMagicLoader : MVRScript
    {	
		private readonly string SKY_GAMEOBJECT_NAME = "SkyMagic Custom Sky";
	
		private JSONStorableFloat myIntensity;
		private JSONStorableBool myUseAsSky;
		private JSONStorableBool mySkyIsHDR;
		private JSONStorableBool mySkyIsSRGB;
		private bool myNeedUpdate = true;
		private bool myHasSetSky = false;
		
		private ReflectionProbe mySkyProbe;
		private JSONStorableFloat mySkyboxYAngle;
		
		public override void Init()
		{
			if (containingAtom.type != "CustomUnityAsset")				
			{
				SuperController.LogError("The MacGruber.SkyMagic plugin needs to be placed on a CustomUnityAsset atom.");
				SetupInfoText(this, 
					"<color=#606060><size=40><b>SkyMagic Loader v0.1</b></size>\n</color><color=#FF0000>SkyMagic plugin needs to be placed on a CustomUnityAsset atom.</color>\n\n",
					1200.0f, true
				);
				myNeedUpdate = false;
				return;
			}			
			
			SetupInfoText(this, 
				"<color=#606060><size=40><b>SkyMagic Loader v0.1</b></size>\nRetrieves a cubemap texture from a ReflectionProbe inside your CustomUnityAsset and sets it as source for VaM's Global Sky and Global Illumination.</color>\n\n" +
				"<b>ReflectionProbe Intensity:</b> Sets the intensity of <b>ALL</b> ReflectionProbes inside your CustomUnityAsset. This affects all nearby Unity assets.\n\n" +
				"<b>Global Sky Enabled:</b> Uses the <b>FIRST</b> ReflectionProbe found inside your CustomUnityAsset as source for Global Sky and Global Illumination, to be used on all VaM native assets.\n\n" +
				"<b>Global Sky is sRGB:</b> Uses gamma colorspace instead of linear. Try to toggle this if your sky is too dark or too bright.\n\n",
				1200.0f, true
			);
			
			myIntensity = SetupSliderFloat(this, "ReflectionProbe Intensity", 1.0f, 0.0f, 10.0f, false);
			myIntensity.setCallbackFunction += (float v) => { UpdateProbes(); };
			
			myUseAsSky = SetupToggle(this, "Global Sky Enabled", false, false);
			myUseAsSky.setCallbackFunction += (bool v) => { UpdateSky(v); };
						
			mySkyIsSRGB = SetupToggle(this, "Global Sky is sRGB", true, false);
			mySkyIsSRGB.setCallbackFunction += (bool v) => { UpdateSky(); };
			
			SetupButton(this, "Force Sync", () => { myNeedUpdate = true; }, false);
			
			myNeedUpdate = true;
		}
		
		private void OnEnable()
		{
			myNeedUpdate = true;
		}
		
		private void OnDisable()
		{			
			UpdateSky(false);
			GameObject skyGO = containingAtom.transform.Find(SKY_GAMEOBJECT_NAME)?.gameObject;
			if (skyGO != null)
				Destroy(skyGO);
		}
		
		private void Update()
		{
			if (mySkyProbe == null && myHasSetSky)
				myNeedUpdate = true;
			
			if (myNeedUpdate)
			{
				bool haveProbes = UpdateProbes();
				bool haveSky = UpdateSky();
				myNeedUpdate = !haveProbes || !haveSky;
			}
			
			if (mySkyProbe != null && mySkyboxYAngle.val != mySkyProbe.transform.eulerAngles.y)
				mySkyboxYAngle.val = mySkyProbe.transform.eulerAngles.y;
		}
		
		private bool UpdateSky()
		{
			return UpdateSky(myUseAsSky.val);
		}
		
		private bool UpdateSky(bool useAsSky)
		{
			SkyshopLightController sslc = SkyshopLightController.singleton;
			Transform t = containingAtom.transform.Find("reParentObject/object/rescaleObject");
			if (t != null && t.childCount > 0 && useAsSky)
			{			
				mySkyProbe = t.GetComponentInChildren<ReflectionProbe>();
				if (mySkyProbe != null)
				{					
					GameObject skyGO = containingAtom.transform.Find(SKY_GAMEOBJECT_NAME)?.gameObject;
					if (skyGO == null)
					{
						skyGO = Instantiate(sslc.skies[0].gameObject, containingAtom.transform);
						skyGO.name = SKY_GAMEOBJECT_NAME;
					}
					Sky sky = skyGO.GetComponent<Sky>();
					sky.SpecularCube = mySkyProbe.customBakedTexture;
					sky.SkyboxCube = mySkyProbe.customBakedTexture;
					sky.HDRSky = false;
					sky.HDRSpec = false;
					sky.LinearSpace = !mySkyIsSRGB.val;
					mySkyboxYAngle = sslc.GetFloatJSONParam("skyboxYAngle");					
					mySkyboxYAngle.setCallbackFunction -= UpdateSkyboxYAngle;
					mySkyboxYAngle.val = mySkyProbe.transform.eulerAngles.y;
					mySkyboxYAngle.setCallbackFunction += UpdateSkyboxYAngle;
					sslc.skyPopup.onValueChangeHandlers -= UpdateSkyboxName;
					sslc.skyPopup.onValueChangeHandlers += UpdateSkyboxName;
					sslc.skyPopup.currentValueNoCallback = SKY_GAMEOBJECT_NAME;					
					sslc.skyManager.BlendToGlobalSky(sky);
					myHasSetSky = true;
					return true;
				}
			}
			
			if (myHasSetSky)
			{
				string skyName = sslc.skyName;
				Sky sky = Array.Find(sslc.skies, s => s.name != skyName);
				sslc.skyName = sky.name;
				sslc.skyPopup.onValueChangeHandlers -= UpdateSkyboxName;
				sslc.skyPopup.currentValue = skyName;
				sslc.skyPopup.onValueChangeHandlers += UpdateSkyboxName;
				mySkyProbe = null;
				myHasSetSky = false;
			}
			
			return false;
		}
		
		private void UpdateSkyboxYAngle(float v)
		{
			// user changed 'Skybox Y Rotation' slider
			if (mySkyProbe != null)
				mySkyProbe.transform.eulerAngles = new Vector3(0, v, 0);
		}
		
		private void UpdateSkyboxName(string skyName)
		{
			// user changed back to one of the default skies
			mySkyProbe = null;
			myHasSetSky = false;
			myUseAsSky.val = false;
		}
		
		private bool UpdateProbes()
		{
			Transform t = containingAtom.transform.Find("reParentObject/object/rescaleObject");
			if (t == null || t.childCount == 0)
				return false;
			
			ReflectionProbe[] probes = t.GetComponentsInChildren<ReflectionProbe>(true);
			for (int i=0; i<probes.Length; ++i)
				probes[i].intensity = myIntensity.val;
			return true;
		}
		
		private static void SetupButton(MVRScript script, string label, UnityAction callback, bool rightSide)
		{
			UIDynamicButton button = script.CreateButton(label, rightSide);
			button.button.onClick.AddListener(callback);
		}
			
		private static JSONStorableBool SetupToggle(MVRScript script, string label, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(label, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateToggle(storable, rightSide);
			script.RegisterBool(storable);
			return storable;
		}
		
		private static JSONStorableFloat SetupSliderFloat(MVRScript script, string label, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateSlider(storable, rightSide);
			script.RegisterFloat(storable);
			return storable;
		}
		
		private static JSONStorableString SetupInfoText(MVRScript script, string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = script.CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
    }
}
