﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
SuperResolution v0.5 by MacGruber.
Makes the VaM internal screenshot tool do screenshots in 8K resolution.

Version 0.5 2020-03-21
	Adding to MacGruber-Toolkit Package.
	
Version 0.4 2019-10-12
	Fixing FoV issue with the preview camera.
	Compatibility with PostMagic.	
	Added override for MSAA setting applied to screenshots.

Version 0.3 2019-10-04
	Fixed VR mode screenshots issue with VaM 1.18

Version 0.2 2019-08-18
	Adding preview camera to allow use in VR mode without performance issues.
	Adding slider with resolutions to select from.

Version 0.1 2019-08-16
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.PostProcessing;
using System.Collections.Generic;
using MeshVR;

namespace MacGruber
{	
	public class SuperResolution : MVRScript
	{	
		private const int READY = 0;
		private const int SCREENSHOT = 1;
		private const int CLEANUP = 2;
		
		private const int resolutionDefaultIdx = 2;		
		private readonly int[] resolutionWidths = new int[] { 9600, 8192, 7680, 5760, 3840 };
		private readonly int[] resolutionHeights = new int[] { 5400, 4608, 4320, 3240, 2160 };
		private readonly string[] resolutionNames = new string[] {
			"9600x5400 (5x FullHD)",
			"8192x4608                   ",
			"7680x4320 (8K UHD)   ",
			"5760x3240 (3x FullHD)",
			"3840x2160 (4K UHD)   "
		};
		
		private readonly int[] msaaLevels = new int[] { 1, 2, 4, 8 };
		private readonly string[] msaaNames = new string[] { "Off", "2x", "4x", "8x" };		
		private int defaultMsaaIdx = 3;
		private int msaaLevel = 8;
		
		private RenderTexture screenshotTexture;
		private Camera previewCamera;
		private Camera screenshotCamera;
		private int state = 0;
		private int resolutionIdx = resolutionDefaultIdx;
		private bool navigationWasDisabled = false;
		
		
		public override void Init()
		{	
			SetupInfoText(this, 
				"<color=#606060><size=40><b>SuperResolution v0.5</b></size>\nTake ultra-high resolution screenshots using VaMs build-in screenshot tool.</color>\n\nSet a resolution here, then go to:\n-> Open Main UI\n-> File (Open/Save)\n-> Screenshot Mode",
				500.0f, true
			);
		
			List<string> resolutionList = new List<string>(resolutionNames);
			JSONStorableStringChooser resolutionChooser = new JSONStorableStringChooser("Resolution", resolutionList, resolutionList[resolutionDefaultIdx], "Resolution");
			resolutionChooser.setCallbackFunction += (string modeName) => {
				resolutionIdx = resolutionList.FindIndex((string entry) => { return entry == modeName; });
				if (resolutionIdx < 0 || resolutionIdx >= resolutionWidths.Length  || resolutionIdx >= resolutionHeights.Length)
					resolutionIdx = resolutionDefaultIdx;
			};
			CreateScrollablePopup(resolutionChooser, false);
			RegisterStringChooser(resolutionChooser);
			
			List<string> msaaList = new List<string>(msaaNames);
			JSONStorableStringChooser msaaChooser = new JSONStorableStringChooser("MSAA Override", msaaList, msaaList[defaultMsaaIdx], "MSAA Override");
			msaaChooser.setCallbackFunction += (string modeName) => {
				int idx = msaaList.FindIndex((string entry) => { return entry == modeName; });
				if (idx < 0 || idx >= msaaLevels.Length)
					idx = defaultMsaaIdx;
				msaaLevel = msaaLevels[idx];
			};
			CreateScrollablePopup(msaaChooser, false);
			RegisterStringChooser(msaaChooser);
			msaaChooser.setCallbackFunction(msaaChooser.val);
		}

		private void OnEnable()
		{
			SuperController sc = SuperController.singleton;
			screenshotCamera = sc.hiResScreenshotCamera;
			previewCamera = Instantiate<Camera>(screenshotCamera, screenshotCamera.transform);
			previewCamera.targetTexture = screenshotCamera.targetTexture;			
			previewCamera.enabled = false;
						
			screenshotCamera.targetTexture = null; // prevent regular VaM screenshot
			screenshotCamera.enabled = false; // prevent expensive rendering in the background
			
			state = READY;
		}
		
		private void Update()
		{	
			SuperController sc = SuperController.singleton;
			GlobalSceneOptions gso = GlobalSceneOptions.singleton;
			
			switch (state)
			{
				case READY:
				{
					// enable cameras as needed
					bool active = sc.hiResScreenshotPreview.gameObject.activeSelf;
					screenshotCamera.enabled = false;
					previewCamera.enabled = active;
					previewCamera.fieldOfView = screenshotCamera.fieldOfView;
					
					if (!active)
						return;
					
					// compatibility with PostMagic
					PostProcessingBehaviour screenshotBehaviour = screenshotCamera.GetComponent<PostProcessingBehaviour>();
					PostProcessingBehaviour previewBehaviour = previewCamera.GetComponent<PostProcessingBehaviour>();
					if (screenshotBehaviour == null)
					{
						if (previewBehaviour != null)
							previewBehaviour.enabled = false;
					}
					else
					{
						if (previewBehaviour == null)
							previewBehaviour = previewCamera.gameObject.AddComponent<PostProcessingBehaviour>();
						previewBehaviour.enabled = true;
						previewBehaviour.profile = screenshotBehaviour.profile;
					}
										
					// ensure RenderTexture size
					if (screenshotTexture == null || screenshotTexture.width != resolutionWidths[resolutionIdx] || screenshotTexture.antiAliasing != msaaLevel)
					{
						if (screenshotTexture != null)
						{
							screenshotTexture.Release();
							Destroy(screenshotTexture);
						}
						RenderTextureDescriptor descriptor = previewCamera.targetTexture.descriptor;			
						descriptor.width = resolutionWidths[resolutionIdx];
						descriptor.height = resolutionHeights[resolutionIdx];
						descriptor.msaaSamples = msaaLevel;
						screenshotTexture = new RenderTexture(descriptor);
						screenshotTexture.Create();
					}
					
					// check for user input
					if (!sc.GetRightSelect() && !sc.GetLeftSelect() && !sc.GetMouseSelect())
						return;
										
					// init screenshot sequence
					screenshotCamera.targetTexture = screenshotTexture;			
					screenshotCamera.enabled = true;
					state = SCREENSHOT;
					break;
				}
				case SCREENSHOT:
				{	
					// fake user input to cause SuperController.ProcessHiResScreenshot do run
					navigationWasDisabled = gso.disableNavigation;
					gso.disableNavigation = true;
					sc.SetLeftSelect(); 
					state = CLEANUP;
					break;
				}
				case CLEANUP:
				{				
					// cleanup
					gso.disableNavigation = navigationWasDisabled;
					screenshotCamera.targetTexture = null;
					screenshotCamera.enabled = false;
					state = READY;
					break;
				}
			}
		}
		
		private void OnDisable()
		{
			if (screenshotTexture != null)
			{
				screenshotTexture.Release();
				Destroy(screenshotTexture);
				screenshotTexture = null;
			}
			
			screenshotCamera.targetTexture = previewCamera.targetTexture;
			screenshotCamera.enabled = SuperController.singleton.hiResScreenshotPreview.gameObject.activeSelf;
			screenshotCamera = null;
			
			Destroy(previewCamera);
			previewCamera = null;
		}
		
		private static JSONStorableString SetupInfoText(MVRScript script, string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = script.CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
	}
}